use crate::consts::K_BACKSPACE;
use crate::rime::Rime;
use crate::utils::{diff, DiffResult};
use regex::Regex;

/// regex raw input to groups
#[derive(Debug)]
pub struct Input<'s> {
    pub raw_text: &'s str,
    pub trigger: Option<&'s str>,
    pub pinyin: &'s str,
    pub select: &'s str,
}

impl<'s> Input<'s> {
    pub fn from_str(re: &Regex, text: &'s str) -> Option<Self> {
        re.captures(text).map(|caps| Input {
            raw_text: caps.get(0).unwrap().as_str(),
            trigger: caps.name("tr").map(|c| c.as_str()),
            pinyin: caps.name("py").unwrap().as_str(),
            select: caps.name("se").unwrap().as_str(),
        })
    }
}

/// which pattern is used for this input
#[derive(Debug)]
pub enum InputKind {
    NoTrigger,
    Trigger,
}

/// cached input state
#[derive(Debug)]
pub struct InputState {
    pub raw_text: String,
    pub session_id: usize,
    pub offset: usize,
    pub kind: InputKind,
}

pub struct InputResult {
    pub is_new: bool,
    pub select: Option<usize>,
}

impl InputState {
    /// check if cached input is prefix or suffix of current input
    /// return diff result: Add / Delete / New
    pub fn new(raw_text: String, session_id: usize, offset: usize, kind: InputKind) -> InputState {
        InputState {
            raw_text,
            session_id,
            offset,
            kind,
        }
    }
    pub fn handle_new_input(
        &self,
        last_input: Input,
        new_offset: usize,
        new_input: &Input,
        rime: &Rime,
    ) -> Result<InputResult, Box<dyn std::error::Error>> {
        // new typing
        if self.offset != new_offset {
            rime.destroy_session(self.session_id);
            return Ok(InputResult {
                is_new: true,
                select: None,
            });
        }
        // continue last typing
        // handle pinyin
        let diff_pinyin = diff(last_input.pinyin, new_input.pinyin);
        match diff_pinyin {
            DiffResult::Add(suffix) => {
                for key in suffix.bytes() {
                    rime.process_key(self.session_id, key as i32);
                }
            }
            DiffResult::Delete(suffix) => {
                for _ in 0..suffix.len() {
                    rime.process_key(self.session_id, K_BACKSPACE);
                }
            }
            DiffResult::New => {
                rime.destroy_session(self.session_id);
            }
            _ => (),
        }
        // handle selection
        let idx = match diff_pinyin {
            DiffResult::Delete(_) => None,
            _ if last_input.select.is_empty() && !new_input.select.is_empty() => {
                Some(new_input.select.parse::<usize>().unwrap())
            }
            _ => None,
        };

        Ok(InputResult {
            is_new: matches!(diff_pinyin, DiffResult::New),
            select: idx,
        })
    }
}
